/**
Die Klasse User dient zur Speicherung von Usern
in einem Online-Shop.

- author: Peter Brichzin
- version: 23.5.24
 */
class User
{
    /** Nachname - Schlüssel dieses Datenelements */
    private var nachname: String
    /** Vorname  */
    private var vorname: String
    /**  Alter  */
    private var alter: Int

    /**
    Konstruktor für Objekte der Klasse User
    - parameters:
        - nachnameNeu: Nachname
        - vornameNeu: Vorname
        - alterNeu: Alter
    */
    init(nachnameNeu: String, vornameNeu: String, alterNeu: Int)
    {
        nachname = nachnameNeu
        vorname = vornameNeu
        alter = alterNeu
    }

    /**
     Gibt den Schlüssel des Datenelements zu Kontrollzwecken
    auf das Terminalfenster aus.
    */
    func InformationAusgeben()
    {
        print(nachname)
    }

    /**
    Gibt alle Information über das Datenelement zu Kontrollzwecken
    auf das Terminalfenster aus.
    */
    func InformationAusgeben2()
    {
        print("Der User " + vorname + " " + nachname + " ist \(alter) Jahre alt.")
    }

    /**
    ergleicht zwei Datenelemente auf Gleichheit.
    - parameters:
        - dvergleich: Datenelement mit dem das Objekt verglichen wird.
    - returns: true, wenn die beiden Datenelemente gleichen Schlüssel haben.
    */
    func IstGleich(dVergleich: User) -> Bool
    {
        return nachname == dVergleich.NachnameGeben()
    }

    /**
    Vergleicht zwei Datenelemente bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichselement: Datenelement mit dem das Objekt verglichen wird.
        - returns: true, wenn das aktuelle Element einen größeren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func IstGrößerAls(dVergleich: User) -> Bool
    {
        return nachname > dVergleich.NachnameGeben()
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, auf Gleichheit.
    - parameters:
        - vergleichsSchlüssel: Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn die beiden Schlüssel gleich sind.
    */
    func SchlüsselIstGleich(vergleichsSchlüssel: String) -> Bool
    {
        return nachname == vergleichsSchlüssel
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichsSchlüssel: Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen grö&szlig;eren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func SchlüsselIstGrößerAls(vergleichsSchlüssel: String) -> Bool
    {
        return nachname > vergleichsSchlüssel
    }

    /**
    Gibt den Schlüssel als Zeichenkette zurück, macht dadurch das Projekt ein wenig übertragbarer.
    - returns: Schlüssel als Zeichenkette
    */
    func SchlüsselAlsStringGeben() -> String
    {
        return nachname
    }

    /**
    Geben-Methode zum Attribut nachname
    - returns: Nachname
    */
    func NachnameGeben() -> String
    {
        return nachname
    }

    /**
    Geben-Methode zum Attribut vorname
    - returns: Vorname
    */
    func VornameGeben() -> String
    {
        return vorname
    }

    /**
    Geben-Methode zum Attribut alter
    - returns: Identifikationsnummer
    */
    func AlterGeben() -> Int
    {
        return alter
    }
}

/**
Abstrakte Klasse Baumelement zur Realisierung einer Baumstruktur mit Hilfe des Entwurfsmusters
Kompositum: Die Klasse legt ausschließlich die Schnittstelle für die Klassen
Knoten und Abschluss in Form von Methoden fest.

- author: Peter Brichzin
- version: 23.5.24
*/
class Baumelement
{
    /**
     Der Konstruktor ist nur der Vollständigkeit halber angegeben.
     */
    init()
    {
        // nichts zu tun
    }
    
    /**
     Fügt ein Datenelement, falls sein Schlüssel noch nicht vorhanden ist,
     in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum
     bleibt erhalten.
     - parameters:
     - datenNeu: das Datenelement des einzufügenden Knoten
     - returns: (neues) Nachfolgeelement des Aufrufers
     */
    func Einfügen(datenNeu: User) -> Baumelement
    {
        fatalError("Baumelement.Einfügen muss überschrieben werden.")
    }
    
    /**
     Sucht ein Datenelement, das über seinen Schlüssel identifiziert wird
     und gibt dieses aus.
     - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
     - returns: gesuchtes Datenelement bzw. nil, falls die Suche erfolglos ist
     */
    func Suchen(suchSchlüssel: String) -> User?
    {
        fatalError("Baumelement.Suchen muss überschrieben werden.")
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true, im Erfolgsfall, false sonst.
     */
    func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        fatalError("Baumelement.IstVorhanden muss überschrieben werden.")
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
     */
    func HöheGeben(aktTiefe: Int) -> Int
    {
        fatalError("Baumelement.HöheGeben muss überschrieben werden.")
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchluessel gespeichert ist.
     - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe die Teife des aktuellen Knotens
    - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
     */
    func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        fatalError("Baumelement.TiefeGeben muss überschrieben werden.")
    }

    /**
    Gibt alle in den Datenelementen eines Baums gespeicherten Informationen Inorder auf der Konsole aus.
     */
    func InorderAusgeben()
    {
        fatalError("Baumelement.InorderAusgeben muss überschrieben werden.")
    }
}

/**
Klasse Abschluss als Bestandteil eines geordneten Binärbaums.
(Implementierung mit dem Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
*/
class Abschluss: Baumelement
{
    /**
     Der Konstruktor ist nur der Vollständigkeit halber angegeben.
     */
    override init()
    {
        super.init()
    }
    
    /**
     Fügt ein Datenelement in die bestehende Datenstruktur ein, indem es einen neuen Knoten
     (mit zwei Abschluss-Objekten als Nachfolger) erzeugt und eine Referenz auf diesen Knoten
     zurückgibt. Das Abschluss-Objekt, dass diese Methode ausführt, wird dabei aus dem Baum entfernt.
     - parameters:
        - datenNeu: das Datenelement des einzufügenden Knoten
     - returns: Referenz auf den neu erzeugten Knoten
     */
    override func Einfügen(datenNeu: User) -> Baumelement
    {
        return Knoten(datenNeu: datenNeu)
    }
    
    /**
     Sucht ein Datenelement, das über seinen Schlüssel identifiziert wird
     und gibt dieses aus.
     - parameters:
     - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
     - returns: nil, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
     */
    override func Suchen(suchSchlüssel: String) -> User?
    {
        return nil
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true, im Erfolgsfall, false sonst.
     */
    override func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        return false
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die übergebene Höhe -1, da hier beim Abschluss das Ende eines Astes erreicht ist und die Abschlusselemente bei der Höhenberechnung nicht einbezogen werden.
     */
    override func HöheGeben(aktTiefe: Int) -> Int
    {
        aktTiefe - 1
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
     - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe die Tiefe des aktuellen Knotens
    - returns: -1 da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
     */
    override func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        return -1
    }

    /**
    Gibt alle in den Datenelementen eines Baums gespeicherten Informationen Inorder auf der Konsole aus.
     */
    override func InorderAusgeben()
    {
        //tue nichts
    }
}

/**
Klasse Knoten zur Verwaltung von Usern als Datenelement
innerhalb eines geordneten Binärbaums.
(Implementierung ohne Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
*/
class Knoten: Baumelement
{

    /** Referenz auf das Datenelement */
    private var daten: User
    /** Referenz auf den linken Nachfolger */
    private var linkerNachfolger: Baumelement
    /** Referenz auf den rechten Nachfolger */
    private var rechterNachfolger: Baumelement

    /**
    Konstruktor für Objekte der Klasse Knoten ohne linken und rechten Nachfolger, d.h. es werden
    zwei Abschluss-Objekte erzeugt und den entsprechenden Referenzattributen zugewiesen.
    - parameters:
        - datenNeu: Referenz auf das Datenelement, das vom Knoten verwaltet wird.
    */
    init(datenNeu: User)
    {
        daten = datenNeu
        linkerNachfolger = Abschluss()
        rechterNachfolger = Abschluss()
        super.init()
    }

    /**
    Fügt ein Datenelement (User), falls sein Schlüssel noch nicht vorhanden ist,
    in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum
    bleibt erhalten.
    - parameters:
        - datenNeu: einzufügender User als Datenelement
     - returns: Referenz auf das eigene Objekt
    */
    override func Einfügen(datenNeu: User) -> Baumelement
    {
        // Falls es ein Datenelement mit dem gleichen Schlüssel schon gibt: Einfügen abbrechen.
        if daten.IstGleich(dVergleich: datenNeu)
        {
            print("Die Daten existieren schon!")
        }
        // Abfrage, ob im rechten oder linken Teilbaum eingefügt werden muss
        else
        {
            if daten.IstGrößerAls(dVergleich: datenNeu)
            {
                linkerNachfolger = linkerNachfolger.Einfügen(datenNeu: datenNeu)
            }
            else
            {
                rechterNachfolger = rechterNachfolger.Einfügen(datenNeu: datenNeu)
            }
        }
        return self
    }

    /**
    Sucht ein Datenelement (User), das über seinen Schlüssel identifiziert wird
    und gibt dieses aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird
    - returns: gesuchtes Datenelement (User) bzw. nil, falls die Suche erfolglos ist
    */
    override func Suchen(suchSchlüssel: String) -> User?
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return daten
        }
        else
        {
            if (daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel))
            {
                return linkerNachfolger.Suchen(suchSchlüssel: suchSchlüssel); // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
            else
            {
                return rechterNachfolger.Suchen(suchSchlüssel: suchSchlüssel); // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
        }
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true, im Erfolgsfall, false sonst.
     */
    override func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return true
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.IstVorhanden(suchSchlüssel: suchSchlüssel) // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
            else
            {
                return rechterNachfolger.IstVorhanden(suchSchlüssel: suchSchlüssel) // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
        }
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
     */
    override func HöheGeben(aktTiefe: Int) -> Int
    {
        // lokale Attriubute
        var maxHöheLinkerTeilbaum: Int
        var maxHöheRechterTeilbaum: Int

        maxHöheLinkerTeilbaum = linkerNachfolger.HöheGeben(aktTiefe: aktTiefe + 1)
        maxHöheRechterTeilbaum = rechterNachfolger.HöheGeben(aktTiefe: aktTiefe + 1)

        if maxHöheLinkerTeilbaum > maxHöheRechterTeilbaum
        {
            return maxHöheLinkerTeilbaum
        }
        else
        {
            return maxHöheRechterTeilbaum
        }
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchluessel gespeichert ist.
     - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe die Teife des aktuellen Knotens
    - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
     */
    override func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return tiefe
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: tiefe + 1)
            }
            else
            {
                return rechterNachfolger.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: tiefe + 1)
            }
        }
    }

    /**
    Gibt alle in den Datenelementen eines Baums gespeicherten Informationen Inorder auf der Konsole aus.
     */
    override func InorderAusgeben()
    {
        linkerNachfolger.InorderAusgeben()
        daten.InformationAusgeben()
        rechterNachfolger.InorderAusgeben()
    }
}

/**
Die Klasse Binbaum ist die Grundstruktur eines geordneten Binärbaums
und hier zur Verwaltung von Usern eingesetzt wird.
Die Methoden die ein Objekt dieser Klasse anbietet, leiten den Aufruf an den
Wurzelknoten weiter, wenn der Baum nicht leer ist.
(Implementierung mit Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 24.5.24
*/
 class BinBaum
{
    /** Die Wurzel des Baums */
     private var wurzel: Baumelement

    /**
    Konstruktor für Objekte der Klasse Binbaum:
    Ein leerer Baum wird erzeugt.
    */
    init()
    {
        wurzel = Abschluss()
    }

    /**
    Fügt ein Datenelement sortiert in den geordneten Binärbaum ein,
    d.h. die Eigenschaft, dass der Baum geordnet ist bleibt bei jedem
    Einfügevorgang erhalten.
    - parameters:
        - datenNeu: neues Datenelement
    */
     func Einfügen(datenNeu: User)
    {
        wurzel = wurzel.Einfügen(datenNeu: datenNeu)
    }

    /**
    Sucht ein Datenelement, das über seinen Schlüssel
    identifiziert wird, und gibt es aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird
    - returns: gesuchtes Datenelement bzw. nil, falls die Suche erfolglos ist
      */
     func Suchen(suchSchlüssel: String) -> User?
     {
         return wurzel.Suchen(suchSchlüssel: suchSchlüssel)
     }

     /**
     Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer Zeichenkette) vorhanden ist.
      - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
      - returns: true, im Erfolgsfall, false sonst.
      */
     func IstVorhanden(suchSchlüssel: String) -> Bool
     {
         return wurzel.IstVorhanden(suchSchlüssel: suchSchlüssel)
     }

     /**
     Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - returns: Höhe des Baums
      */
     func HöheGeben() -> Int
     {
         return wurzel.HöheGeben(aktTiefe: 0)
     }

     /**
     Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
      - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
      - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
      */
     func TiefeGeben(suchSchlüssel: String) -> Int
     {
         return wurzel.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: 0)
     }

     /**
     Gibt den Baum Inorder als Liste auf der Konsole aus.
      */
     func InorderAusgeben()
     {
         wurzel.InorderAusgeben()
     }
}

let user1 = User(nachnameNeu: "Geber", vornameNeu: "Ann", alterNeu: 16)
let user2 = User(nachnameNeu: "Tross", vornameNeu: "Albert", alterNeu: 16)
let user3 = User(nachnameNeu: "Fall", vornameNeu: "Klara", alterNeu: 17)
let user4 = User(nachnameNeu: "Ter", vornameNeu: "Lee", alterNeu: 18)
let user5 = User(nachnameNeu: "Sehr", vornameNeu: "Kai", alterNeu: 20)
let user6 = User(nachnameNeu: "Pfahl", vornameNeu: "Marta", alterNeu: 17)
let  binBaum1 = BinBaum()
binBaum1.Einfügen(datenNeu: user5)
binBaum1.Einfügen(datenNeu: user1)
binBaum1.Einfügen(datenNeu: user2)
binBaum1.Einfügen(datenNeu: user3)
binBaum1.Einfügen(datenNeu: user4)
binBaum1.Einfügen(datenNeu: user6)
binBaum1.InorderAusgeben()
