/**
Die Klasse Wörterbucheintrag dient zur Speicherung von
Wörterbucheinträgen für ein Englisch-Deutsches
Wörterbuch. Dabei kann jedem englischen Wort mehrere
deutsche Bedeutungen zugeordnet werden, wobei alle Bedeutungen in
einer einzigen Zeichenkette Bedeutung gespeichert werden.

- author: Peter Brichzin
- version: 23.05.24
*/
class Wörterbucheintrag
{
    /** Englisches Wort;  Schlüssel dieses Datenelements */
    private var wort: String
    /** Deutsche Bedeutung; mehrere Bedeutungen werden als eine Zeichenkette zusammengefasst */
    private var bedeutung: String

    /**
    Konstruktor für Objekte der Klasse Wörterbucheintrag
    - parameters:
        - wortNeu: das englische Wort.
        - bedeutungNeu die deutsche(n) Bedeutung(en).
    */
    init(wortNeu: String, bedeutungNeu: String)
    {
        wort = wortNeu
        bedeutung = bedeutungNeu
    }

    /**
    Gibt Information über das Datenelement zu Kontrollzwecken
    auf das Terminalfenster aus.
    */
    func InformationAusgeben()
    {
        print(wort + ": " + bedeutung)
    }

    /**
    Vergleicht zwei Datenelemente auf Gleichheit.
    - parameters:
        - dvergleich: Datenelement mit dem das Objekt verglichen wird.
    - returns: true, wenn die beiden Datenelemente gleichen Schlüssel haben.
    */
    func IstGleich(dvergleich: Wörterbucheintrag) -> Bool
    {
        if wort == dvergleich.WortGeben()
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Vergleicht zwei Datenelemente bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichselement: Datenelement mit dem das Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen größeren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func IstGrößerAls(dvergleich: Wörterbucheintrag) -> Bool
    {
        if wort > dvergleich.WortGeben()
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, auf Gleichheit.
    - parameters:
        - vergleichsSchlüssel: Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn die beiden Schlüssel gleich sind.
    */
    func SchlüsselIstGleich(vergleichsSchlüssel: String) -> Bool
    {
        return wort == vergleichsSchlüssel
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichsSchlüssel Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen größeren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func SchlüsselIstGrößerAls(vergleichsSchlüssel: String) -> Bool
    {
        if wort > vergleichsSchlüssel
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichsSchüssel Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen kleineren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func SchlüsselIstKleinerAls(vergleichsSchlüssel: String) -> Bool
    {
        if wort < vergleichsSchlüssel
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Gibt den Schlüssel des Datenelements als String aus, auch wenn der Datentyp vom
    String abweicht. Diese Methode wird beim Entfernen eines Knotens benötigt.
    - returns: Schlüssel als String.
     */
    func SchlüsselAlsStringGeben() -> String
    {
        return wort
    }

    /**
    Geben-Methode zum Attribut wort
    - returns: das englische Wort
    */
    func WortGeben() -> String
    {
        return wort
    }

    /**
    Geben-Methode zum Attribut bedeutung
    - returns: bedeutung: die deutsche(n) Bedeutung(en) als eine Zeichenkette
    */
    func BedeutungGeben() -> String
    {
        return bedeutung
    }

    /**
    Setzen-Methode zum Attribut bedeutung
    - parameters:
        - bedeutungNeu: neue Bedeutung für das englische Wort
    */
    func BedeutungSetzen(bedeutungNeu: String)
    {
        bedeutung = bedeutungNeu
    }
}

/**
Abstrakte Klasse Baumelement zur Realisierung einer Baumstruktur mit Hilfe des Entwurfsmusters
Kompositum: Die Klasse legt ausschließlich die Schnittstelle für die Klassen
Knoten und Abschluss in Form von Methoden fest.

- author: Peter Brichzin
- version: 23.5.24
*/
class Baumelement
{
    /**
    Der Konstruktor ist nur der Vollständigkeit halber angegeben.
    */
    init()
    {
        // nichts zu tun
    }

    /**
    Fügt ein Datenelement, falls sein Schlüssel noch nicht vorhanden ist,
    in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum
    bleibt erhalten.
    - parameters:
        - datenNeu: das Datenelement des einzufügenden Knoten
    - returns: (neues) Nachfolgeelement des Aufrufers
    */
    func Einfügen(datenNeu: Wörterbucheintrag) -> Baumelement
    {
        fatalError("Baumelement.Einfügen muss überschrieben werden.")
    }

    /**
    Sucht ein Datenelement, das über seinen Schlüssel identifiziert wird
    und gibt dieses aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
    - returns: gesuchtes Datenelement bzw. nil, falls die Suche erfolglos ist
    */
    func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        fatalError("Baumelement.Suchen muss überschrieben werden.")
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true im Erfolgsfall, false sonst.
    */
    func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        fatalError("Baumelement.IstVorhanden muss überschrieben werden.")
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
    */
    func HöheGeben(aktTiefe: Int) -> Int
    {
        fatalError("Baumelement.HöheGeben muss überschrieben werden.")
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe: die Tiefe des aktuellen Knotens
    - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
    */
    func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        fatalError("Baumelement.TiefeGeben muss überschrieben werden.")
    }
}

/**
Klasse Abschluss als Bestandteil eines geordneten Binärbaums.
(Implementierung mit dem Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
*/
class Abschluss: Baumelement
{
    /**
    Der Konstruktor ist nur der Vollständigkeit halber angegeben.
    */
    override init()
    {
        super.init()
    }

    /**
    Fügt ein Datenelement in die bestehende Datenstruktur ein, indem es einen neuen Knoten
    (mit zwei Abschluss-Objekten als Nachfolger) erzeugt und eine Referenz auf diesen Knoten
    zurückgibt. Das Abschluss-Objekt, dass diese Methode ausführt, wird dabei aus dem Baum entfernt.
    - parameters:
        - datenNeu: das Datenelement des einzufügenden Knoten
    - returns: Referenz auf den neu erzeugten Knoten
    */
    override func Einfügen(datenNeu: Wörterbucheintrag) -> Baumelement
    {
        return Knoten(datenNeu: datenNeu)
    }

    /**
    Sucht ein Datenelement, das über seinen Schlüssel identifiziert wird
    und gibt dieses aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
    - returns: nl, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
    */
    override func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        return nil
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: false, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
    */
    override func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        return false
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die übergebene Höhe -1, da hier beim Abschluss das Ende eines Astes erreicht ist und die Abschlusselemente bei der Höhenberechnung nicht einbezogen werden.
    */
    override func HöheGeben(aktTiefe: Int) -> Int
    {
        return aktTiefe - 1
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe: die Tiefe des aktuellen Knotens
    - returns: -1 da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
    */
    override func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        return -1
    }
}

/**
Klasse Knoten als Bestandteil eines geordneten Binärbaums.
(Implementierung mit dem Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
*/
class Knoten: Baumelement
{

    /** Datenelement des Knotens */
    private var daten: Wörterbucheintrag
    /** Referenz auf den linken Nachfolger */
    private var linkerNachfolger: Baumelement
    /** Referenz auf den rechten Nachfolger */
    private var rechterNachfolger: Baumelement

    /**
     * Konstruktor für Objekte der Klasse Knoten ohne linken und rechten Nachfolger, d.h. es werden
     * zwei Abschluss-Objekte erzeugt und den entsprechenden Referenzattributen zugewiesen.
     * @param datenNeu Referenz auf ein Objekt der KLasse Wörterbucheintrag, das vom Knoten verwaltet wird.
     */
    init(datenNeu: Wörterbucheintrag)
    {
        daten = datenNeu;
        linkerNachfolger = Abschluss()
        rechterNachfolger = Abschluss()
        super.init()
    }

    /**
    Fügt ein Datenelement (Wörterbucheintrag), falls sein Schlüssel noch nicht vorhanden ist,
    in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum
    bleibt erhalten.
    - parameters:
        - datenNeu: einzufügendes Datenelement (Wörterbucheintrag)
    - returns: Referenz auf das eigene Objekt
    */
    override func Einfügen(datenNeu: Wörterbucheintrag) -> Baumelement
    {
        // Falls es den Wert schon gibt, Einfügen abbrechen.
        if daten.IstGleich(dvergleich: datenNeu)
        {
            print("Die Daten existieren schon!")
        }
        // Abfrage ob im rechten oder linken Teilbaum eingefügt werden muss
        else
        {
            if daten.IstGrößerAls(dvergleich: datenNeu)
            {
                linkerNachfolger = linkerNachfolger.Einfügen(datenNeu: datenNeu)
            }
            else
            {
                rechterNachfolger = rechterNachfolger.Einfügen(datenNeu: datenNeu)
            }
        }
        return self
    }

    /**
    Sucht ein Datenelement (Wörterbucheintrag), das über seinen Schlüssel identifiziert wird
    und gibt dieses aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
    - returns: gesuchtes Datenelement (Wörterbucheintrag) bzw. nil, falls die Suche erfolglos ist
    */
    override func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return daten
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.Suchen(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
            else
            {
                return rechterNachfolger.Suchen(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
        }
    }

    /**
    Überprüft, ob ein Datenelement (Wörterbucheintrag) mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true im Erfolgsfall, false sonst.
    */
    override func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return true
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.IstVorhanden(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
            else
            {
                return rechterNachfolger.IstVorhanden(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
        }
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
    */
    override func HöheGeben(aktTiefe: Int) -> Int
    {
        // lokale Attriubute
        let maxHöheLinkerTeilbaum = linkerNachfolger.HöheGeben(aktTiefe: aktTiefe + 1)
        let maxHöheRechterTeilbaum = rechterNachfolger.HöheGeben(aktTiefe: aktTiefe + 1)

        if maxHöheLinkerTeilbaum > maxHöheRechterTeilbaum
        {
            return maxHöheLinkerTeilbaum
        }
        else
        {
            return maxHöheRechterTeilbaum
        }
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe: die Tiefe des aktuellen Knotens
    - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
    */
    override func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return tiefe
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: tiefe + 1)
            }
            else
            {
                return rechterNachfolger.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: tiefe + 1)
            }
        }
    }
}

/**
Die Klasse Binbaum ist die Grundstruktur eines geordneten Binärbaums.
Konkret wird hier der Binärbaum genutzt um ein einfaches Wörterbuch umzusetzen,
deshalb sind die Daten von der Klasse Wörterbucheintrag.
Die Methoden die ein Objekt dieser Klasse anbietet, leiten den Aufruf an den
Wurzelknoten weiter, wenn der Baum nicht leer ist.
(Implementierung mit Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
 */
class BinBaum
{
    /** Die Wurzel des Baums */
    private var wurzel: Baumelement

    /**
    Konstruktor für Objekte der Klasse Binbaum
    Ein leerer Baum wird erzeugt: Er besteht nur aus einem Abschluss-Objekt.
    */
    init()
    {
        wurzel = Abschluss()
    }

    /**
    Fügt ein Datenelement sortiert in den geordneten Binärbaum ein,
    d.h. die Eigenschaft, dass der Baum geordnet ist bleibt bei jedem
    Einfügevorgang erhalten.
    - parameters:
        - datenNeu: neues Datenelement
    */
    func Einfügen(datenNeu: Wörterbucheintrag)
    {
        wurzel = wurzel.Einfügen(datenNeu: datenNeu)
    }

    /**
    Sucht ein Datenelement, das über seinen Schlüssel
    identifiziert wird, und gibt es aus.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird
    - returns: gesuchtes Datenelement bzw. nil, falls die Suche erfolglos ist
    */
    func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        return wurzel.Suchen(suchSchlüssel: suchSchlüssel)
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true, im Erfolgsfall, false sonst.
    */
    func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        return wurzel.IstVorhanden(suchSchlüssel: suchSchlüssel)
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - returns: Höhe des Baums
    */
    func HöheGeben() -> Int
    {
        return wurzel.HöheGeben(aktTiefe: 0)
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
    */
    func TiefeGeben(suchSchlüssel: String) -> Int
    {
        return wurzel.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: 0)
    }
}

let woerterb1 = Wörterbucheintrag(wortNeu: "clip", bedeutungNeu: "Clip, Spange, Klammer")
let woerterb2 = Wörterbucheintrag(wortNeu: "cat", bedeutungNeu: "Katze")
let woerterb3 = Wörterbucheintrag(wortNeu: "cup", bedeutungNeu: "Tasse, Becher")
let woerterb4 = Wörterbucheintrag(wortNeu: "cut", bedeutungNeu: "Schnitt, Öffnung, kürzen")
let woerterb5 = Wörterbucheintrag(wortNeu: "cube", bedeutungNeu: "Würfel")
let woerterb6 = Wörterbucheintrag(wortNeu: "cave", bedeutungNeu: "Höhle")
let binBaum1 = BinBaum()
binBaum1.Einfügen(datenNeu: woerterb1)
binBaum1.Einfügen(datenNeu: woerterb2)
binBaum1.Einfügen(datenNeu: woerterb3)
binBaum1.Einfügen(datenNeu: woerterb4)
let woerterb7 = Wörterbucheintrag(wortNeu: "crab", bedeutungNeu: "Krabbe")
